import * as ValueGuard from './value.mjs';
import { Kind, Hint, TransformKind, ReadonlyKind, OptionalKind } from '../symbols/index.mjs';
export class TypeGuardUnknownTypeError extends Error {
}
const KnownTypes = [
    'Any',
    'Array',
    'AsyncIterator',
    'BigInt',
    'Boolean',
    'Constructor',
    'Date',
    'Enum',
    'Function',
    'Integer',
    'Intersect',
    'Iterator',
    'Literal',
    'Not',
    'Null',
    'Number',
    'Object',
    'Promise',
    'Record',
    'Ref',
    'String',
    'Symbol',
    'TemplateLiteral',
    'This',
    'Tuple',
    'Undefined',
    'Union',
    'Uint8Array',
    'Unknown',
    'Void',
];
function IsPattern(value) {
    try {
        new RegExp(value);
        return true;
    }
    catch {
        return false;
    }
}
function IsControlCharacterFree(value) {
    if (!ValueGuard.IsString(value))
        return false;
    for (let i = 0; i < value.length; i++) {
        const code = value.charCodeAt(i);
        if ((code >= 7 && code <= 13) || code === 27 || code === 127) {
            return false;
        }
    }
    return true;
}
function IsAdditionalProperties(value) {
    return IsOptionalBoolean(value) || TSchema(value);
}
function IsOptionalBigInt(value) {
    return ValueGuard.IsUndefined(value) || ValueGuard.IsBigInt(value);
}
function IsOptionalNumber(value) {
    return ValueGuard.IsUndefined(value) || ValueGuard.IsNumber(value);
}
function IsOptionalBoolean(value) {
    return ValueGuard.IsUndefined(value) || ValueGuard.IsBoolean(value);
}
function IsOptionalString(value) {
    return ValueGuard.IsUndefined(value) || ValueGuard.IsString(value);
}
function IsOptionalPattern(value) {
    return ValueGuard.IsUndefined(value) || (ValueGuard.IsString(value) && IsControlCharacterFree(value) && IsPattern(value));
}
function IsOptionalFormat(value) {
    return ValueGuard.IsUndefined(value) || (ValueGuard.IsString(value) && IsControlCharacterFree(value));
}
function IsOptionalSchema(value) {
    return ValueGuard.IsUndefined(value) || TSchema(value);
}
export function TReadonly(schema) {
    return ValueGuard.IsObject(schema) && schema[ReadonlyKind] === 'Readonly';
}
export function TOptional(schema) {
    return ValueGuard.IsObject(schema) && schema[OptionalKind] === 'Optional';
}
export function TAny(schema) {
    return (TKindOf(schema, 'Any') &&
        IsOptionalString(schema.$id));
}
export function TArray(schema) {
    return (TKindOf(schema, 'Array') &&
        schema.type === 'array' &&
        IsOptionalString(schema.$id) &&
        TSchema(schema.items) &&
        IsOptionalNumber(schema.minItems) &&
        IsOptionalNumber(schema.maxItems) &&
        IsOptionalBoolean(schema.uniqueItems) &&
        IsOptionalSchema(schema.contains) &&
        IsOptionalNumber(schema.minContains) &&
        IsOptionalNumber(schema.maxContains));
}
export function TAsyncIterator(schema) {
    return (TKindOf(schema, 'AsyncIterator') &&
        schema.type === 'AsyncIterator' &&
        IsOptionalString(schema.$id) &&
        TSchema(schema.items));
}
export function TBigInt(schema) {
    return (TKindOf(schema, 'BigInt') &&
        schema.type === 'bigint' &&
        IsOptionalString(schema.$id) &&
        IsOptionalBigInt(schema.exclusiveMaximum) &&
        IsOptionalBigInt(schema.exclusiveMinimum) &&
        IsOptionalBigInt(schema.maximum) &&
        IsOptionalBigInt(schema.minimum) &&
        IsOptionalBigInt(schema.multipleOf));
}
export function TBoolean(schema) {
    return (TKindOf(schema, 'Boolean') &&
        schema.type === 'boolean' &&
        IsOptionalString(schema.$id));
}
export function TConstructor(schema) {
    return (TKindOf(schema, 'Constructor') &&
        schema.type === 'Constructor' &&
        IsOptionalString(schema.$id) &&
        ValueGuard.IsArray(schema.parameters) &&
        schema.parameters.every(schema => TSchema(schema)) &&
        TSchema(schema.returns));
}
export function TDate(schema) {
    return (TKindOf(schema, 'Date') &&
        schema.type === 'Date' &&
        IsOptionalString(schema.$id) &&
        IsOptionalNumber(schema.exclusiveMaximumTimestamp) &&
        IsOptionalNumber(schema.exclusiveMinimumTimestamp) &&
        IsOptionalNumber(schema.maximumTimestamp) &&
        IsOptionalNumber(schema.minimumTimestamp) &&
        IsOptionalNumber(schema.multipleOfTimestamp));
}
export function TFunction(schema) {
    return (TKindOf(schema, 'Function') &&
        schema.type === 'Function' &&
        IsOptionalString(schema.$id) &&
        ValueGuard.IsArray(schema.parameters) &&
        schema.parameters.every(schema => TSchema(schema)) &&
        TSchema(schema.returns));
}
export function TInteger(schema) {
    return (TKindOf(schema, 'Integer') &&
        schema.type === 'integer' &&
        IsOptionalString(schema.$id) &&
        IsOptionalNumber(schema.exclusiveMaximum) &&
        IsOptionalNumber(schema.exclusiveMinimum) &&
        IsOptionalNumber(schema.maximum) &&
        IsOptionalNumber(schema.minimum) &&
        IsOptionalNumber(schema.multipleOf));
}
export function TIntersect(schema) {
    return (TKindOf(schema, 'Intersect') &&
        (ValueGuard.IsString(schema.type) && schema.type !== 'object' ? false : true) &&
        ValueGuard.IsArray(schema.allOf) &&
        schema.allOf.every(schema => TSchema(schema) && !TTransform(schema)) &&
        IsOptionalString(schema.type) &&
        (IsOptionalBoolean(schema.unevaluatedProperties) || IsOptionalSchema(schema.unevaluatedProperties)) &&
        IsOptionalString(schema.$id));
}
export function TIterator(schema) {
    return (TKindOf(schema, 'Iterator') &&
        schema.type === 'Iterator' &&
        IsOptionalString(schema.$id) &&
        TSchema(schema.items));
}
export function TKindOf(schema, kind) {
    return ValueGuard.IsObject(schema) && Kind in schema && schema[Kind] === kind;
}
export function TLiteralString(schema) {
    return TLiteral(schema) && ValueGuard.IsString(schema.const);
}
export function TLiteralNumber(schema) {
    return TLiteral(schema) && ValueGuard.IsNumber(schema.const);
}
export function TLiteralBoolean(schema) {
    return TLiteral(schema) && ValueGuard.IsBoolean(schema.const);
}
export function TLiteral(schema) {
    return (TKindOf(schema, 'Literal') &&
        IsOptionalString(schema.$id) && (ValueGuard.IsBoolean(schema.const) ||
        ValueGuard.IsNumber(schema.const) ||
        ValueGuard.IsString(schema.const)));
}
export function TNever(schema) {
    return (TKindOf(schema, 'Never') &&
        ValueGuard.IsObject(schema.not) &&
        Object.getOwnPropertyNames(schema.not).length === 0);
}
export function TNot(schema) {
    return (TKindOf(schema, 'Not') &&
        TSchema(schema.not));
}
export function TNull(schema) {
    return (TKindOf(schema, 'Null') &&
        schema.type === 'null' &&
        IsOptionalString(schema.$id));
}
export function TNumber(schema) {
    return (TKindOf(schema, 'Number') &&
        schema.type === 'number' &&
        IsOptionalString(schema.$id) &&
        IsOptionalNumber(schema.exclusiveMaximum) &&
        IsOptionalNumber(schema.exclusiveMinimum) &&
        IsOptionalNumber(schema.maximum) &&
        IsOptionalNumber(schema.minimum) &&
        IsOptionalNumber(schema.multipleOf));
}
export function TObject(schema) {
    return (TKindOf(schema, 'Object') &&
        schema.type === 'object' &&
        IsOptionalString(schema.$id) &&
        ValueGuard.IsObject(schema.properties) &&
        Object.entries(schema.properties).every(([key, schema]) => IsControlCharacterFree(key) && TSchema(schema)) &&
        IsAdditionalProperties(schema.additionalProperties) &&
        IsOptionalNumber(schema.minProperties) &&
        IsOptionalNumber(schema.maxProperties));
}
export function TPromise(schema) {
    return (TKindOf(schema, 'Promise') &&
        schema.type === 'Promise' &&
        IsOptionalString(schema.$id) &&
        TSchema(schema.item));
}
export function TRecord(schema) {
    return (TKindOf(schema, 'Record') &&
        schema.type === 'object' &&
        IsOptionalString(schema.$id) &&
        IsAdditionalProperties(schema.additionalProperties) &&
        ValueGuard.IsObject(schema.patternProperties) &&
        ((schema) => {
            const keys = Object.getOwnPropertyNames(schema.patternProperties);
            return (keys.length === 1 &&
                IsPattern(keys[0]) &&
                ValueGuard.IsObject(schema.patternProperties) &&
                TSchema(schema.patternProperties[keys[0]]));
        })(schema));
}
export function TRecursive(schema) {
    return ValueGuard.IsObject(schema) && Hint in schema && schema[Hint] === 'Recursive';
}
export function TRef(schema) {
    return (TKindOf(schema, 'Ref') &&
        IsOptionalString(schema.$id) &&
        ValueGuard.IsString(schema.$ref));
}
export function TString(schema) {
    return (TKindOf(schema, 'String') &&
        schema.type === 'string' &&
        IsOptionalString(schema.$id) &&
        IsOptionalNumber(schema.minLength) &&
        IsOptionalNumber(schema.maxLength) &&
        IsOptionalPattern(schema.pattern) &&
        IsOptionalFormat(schema.format));
}
export function TSymbol(schema) {
    return (TKindOf(schema, 'Symbol') &&
        schema.type === 'symbol' &&
        IsOptionalString(schema.$id));
}
export function TTemplateLiteral(schema) {
    return (TKindOf(schema, 'TemplateLiteral') &&
        schema.type === 'string' &&
        ValueGuard.IsString(schema.pattern) &&
        schema.pattern[0] === '^' &&
        schema.pattern[schema.pattern.length - 1] === '$');
}
export function TThis(schema) {
    return (TKindOf(schema, 'This') &&
        IsOptionalString(schema.$id) &&
        ValueGuard.IsString(schema.$ref));
}
export function TTransform(schema) {
    return ValueGuard.IsObject(schema) && TransformKind in schema;
}
export function TTuple(schema) {
    return (TKindOf(schema, 'Tuple') &&
        schema.type === 'array' &&
        IsOptionalString(schema.$id) &&
        ValueGuard.IsNumber(schema.minItems) &&
        ValueGuard.IsNumber(schema.maxItems) &&
        schema.minItems === schema.maxItems &&
        ((ValueGuard.IsUndefined(schema.items) &&
            ValueGuard.IsUndefined(schema.additionalItems) &&
            schema.minItems === 0) || (ValueGuard.IsArray(schema.items) &&
            schema.items.every(schema => TSchema(schema)))));
}
export function TUndefined(schema) {
    return (TKindOf(schema, 'Undefined') &&
        schema.type === 'undefined' &&
        IsOptionalString(schema.$id));
}
export function TUnionLiteral(schema) {
    return TUnion(schema) && schema.anyOf.every((schema) => TLiteralString(schema) || TLiteralNumber(schema));
}
export function TUnion(schema) {
    return (TKindOf(schema, 'Union') &&
        IsOptionalString(schema.$id) &&
        ValueGuard.IsObject(schema) &&
        ValueGuard.IsArray(schema.anyOf) &&
        schema.anyOf.every(schema => TSchema(schema)));
}
export function TUint8Array(schema) {
    return (TKindOf(schema, 'Uint8Array') &&
        schema.type === 'Uint8Array' &&
        IsOptionalString(schema.$id) &&
        IsOptionalNumber(schema.minByteLength) &&
        IsOptionalNumber(schema.maxByteLength));
}
export function TUnknown(schema) {
    return (TKindOf(schema, 'Unknown') &&
        IsOptionalString(schema.$id));
}
export function TUnsafe(schema) {
    return TKindOf(schema, 'Unsafe');
}
export function TVoid(schema) {
    return (TKindOf(schema, 'Void') &&
        schema.type === 'void' &&
        IsOptionalString(schema.$id));
}
export function TKind(schema) {
    return ValueGuard.IsObject(schema) && Kind in schema && ValueGuard.IsString(schema[Kind]) && !KnownTypes.includes(schema[Kind]);
}
export function TSchema(schema) {
    return (ValueGuard.IsObject(schema)) && (TAny(schema) ||
        TArray(schema) ||
        TBoolean(schema) ||
        TBigInt(schema) ||
        TAsyncIterator(schema) ||
        TConstructor(schema) ||
        TDate(schema) ||
        TFunction(schema) ||
        TInteger(schema) ||
        TIntersect(schema) ||
        TIterator(schema) ||
        TLiteral(schema) ||
        TNever(schema) ||
        TNot(schema) ||
        TNull(schema) ||
        TNumber(schema) ||
        TObject(schema) ||
        TPromise(schema) ||
        TRecord(schema) ||
        TRef(schema) ||
        TString(schema) ||
        TSymbol(schema) ||
        TTemplateLiteral(schema) ||
        TThis(schema) ||
        TTuple(schema) ||
        TUndefined(schema) ||
        TUnion(schema) ||
        TUint8Array(schema) ||
        TUnknown(schema) ||
        TUnsafe(schema) ||
        TVoid(schema) ||
        TKind(schema));
}
