import { TTransform as IsTransformType, TSchema as IsSchemaType } from '../../type/guard/type.mjs';
import { Kind, TransformKind } from '../../type/symbols/index.mjs';
import { IsPlainObject, IsArray, IsValueType } from '../guard/index.mjs';
import { KeyOfStringResolve } from '../../type/keyof/index.mjs';
import { IndexedTypeResolve } from '../../type/indexed/index.mjs';
import { Deref } from '../deref/index.mjs';
import { Check } from '../check/index.mjs';
export class TransformDecodeCheckError extends Error {
    schema;
    value;
    error;
    constructor(schema, value, error) {
        super(`Unable to decode due to invalid value`);
        this.schema = schema;
        this.value = value;
        this.error = error;
    }
}
export class TransformDecodeError extends Error {
    schema;
    value;
    constructor(schema, value, error) {
        super(`${error instanceof Error ? error.message : 'Unknown error'}`);
        this.schema = schema;
        this.value = value;
    }
}
function Default(schema, value) {
    try {
        return IsTransformType(schema) ? schema[TransformKind].Decode(value) : value;
    }
    catch (error) {
        throw new TransformDecodeError(schema, value, error);
    }
}
function TArray(schema, references, value) {
    return (IsArray(value))
        ? Default(schema, value.map((value) => Visit(schema.items, references, value)))
        : Default(schema, value);
}
function TIntersect(schema, references, value) {
    if (!IsPlainObject(value) || IsValueType(value))
        return Default(schema, value);
    const knownKeys = KeyOfStringResolve(schema);
    const knownProperties = knownKeys.reduce((value, key) => {
        return (key in value)
            ? { ...value, [key]: Visit(IndexedTypeResolve(schema, [key]), references, value[key]) }
            : value;
    }, value);
    if (!IsTransformType(schema.unevaluatedProperties)) {
        return Default(schema, knownProperties);
    }
    const unknownKeys = Object.getOwnPropertyNames(knownProperties);
    const unevaluatedProperties = schema.unevaluatedProperties;
    const unknownProperties = unknownKeys.reduce((value, key) => {
        return !knownKeys.includes(key)
            ? { ...value, [key]: Default(unevaluatedProperties, value[key]) }
            : value;
    }, knownProperties);
    return Default(schema, unknownProperties);
}
function TNot(schema, references, value) {
    return Default(schema, Visit(schema.not, references, value));
}
function TObject(schema, references, value) {
    if (!IsPlainObject(value))
        return Default(schema, value);
    const knownKeys = KeyOfStringResolve(schema);
    const knownProperties = knownKeys.reduce((value, key) => {
        return (key in value)
            ? { ...value, [key]: Visit(schema.properties[key], references, value[key]) }
            : value;
    }, value);
    if (!IsSchemaType(schema.additionalProperties)) {
        return Default(schema, knownProperties);
    }
    const unknownKeys = Object.getOwnPropertyNames(knownProperties);
    const additionalProperties = schema.additionalProperties;
    const unknownProperties = unknownKeys.reduce((value, key) => {
        return !knownKeys.includes(key)
            ? { ...value, [key]: Default(additionalProperties, value[key]) }
            : value;
    }, knownProperties);
    return Default(schema, unknownProperties);
}
function TRecord(schema, references, value) {
    if (!IsPlainObject(value))
        return Default(schema, value);
    const pattern = Object.getOwnPropertyNames(schema.patternProperties)[0];
    const knownKeys = new RegExp(pattern);
    const knownProperties = Object.getOwnPropertyNames(value).reduce((value, key) => {
        return knownKeys.test(key)
            ? { ...value, [key]: Visit(schema.patternProperties[pattern], references, value[key]) }
            : value;
    }, value);
    if (!IsSchemaType(schema.additionalProperties)) {
        return Default(schema, knownProperties);
    }
    const unknownKeys = Object.getOwnPropertyNames(knownProperties);
    const additionalProperties = schema.additionalProperties;
    const unknownProperties = unknownKeys.reduce((value, key) => {
        return !knownKeys.test(key)
            ? { ...value, [key]: Default(additionalProperties, value[key]) }
            : value;
    }, knownProperties);
    return Default(schema, unknownProperties);
}
function TRef(schema, references, value) {
    const target = Deref(schema, references);
    return Default(schema, Visit(target, references, value));
}
function TThis(schema, references, value) {
    const target = Deref(schema, references);
    return Default(schema, Visit(target, references, value));
}
function TTuple(schema, references, value) {
    return (IsArray(value) && IsArray(schema.items))
        ? Default(schema, schema.items.map((schema, index) => Visit(schema, references, value[index])))
        : Default(schema, value);
}
function TUnion(schema, references, value) {
    for (const subschema of schema.anyOf) {
        if (!Check(subschema, references, value))
            continue;
        const decoded = Visit(subschema, references, value);
        return Default(schema, decoded);
    }
    return Default(schema, value);
}
function Visit(schema, references, value) {
    const references_ = typeof schema.$id === 'string' ? [...references, schema] : references;
    const schema_ = schema;
    switch (schema[Kind]) {
        case 'Array':
            return TArray(schema_, references_, value);
        case 'Intersect':
            return TIntersect(schema_, references_, value);
        case 'Not':
            return TNot(schema_, references_, value);
        case 'Object':
            return TObject(schema_, references_, value);
        case 'Record':
            return TRecord(schema_, references_, value);
        case 'Ref':
            return TRef(schema_, references_, value);
        case 'Symbol':
            return Default(schema_, value);
        case 'This':
            return TThis(schema_, references_, value);
        case 'Tuple':
            return TTuple(schema_, references_, value);
        case 'Union':
            return TUnion(schema_, references_, value);
        default:
            return Default(schema_, value);
    }
}
export function Decode(schema, references, value) {
    return Visit(schema, references, value);
}
