import { TTransform as IsTransformType, TSchema as IsSchemaType } from '../../type/guard/type.mjs';
import { Kind, TransformKind } from '../../type/symbols/index.mjs';
import { IsPlainObject, IsArray, IsValueType } from '../guard/index.mjs';
import { KeyOfStringResolve } from '../../type/keyof/index.mjs';
import { IndexedTypeResolve } from '../../type/indexed/index.mjs';
import { Deref } from '../deref/index.mjs';
import { Check } from '../check/index.mjs';
export class TransformEncodeCheckError extends Error {
    schema;
    value;
    error;
    constructor(schema, value, error) {
        super(`Unable to encode due to invalid value`);
        this.schema = schema;
        this.value = value;
        this.error = error;
    }
}
export class TransformEncodeError extends Error {
    schema;
    value;
    constructor(schema, value, error) {
        super(`${error instanceof Error ? error.message : 'Unknown error'}`);
        this.schema = schema;
        this.value = value;
    }
}
function Default(schema, value) {
    try {
        return IsTransformType(schema) ? schema[TransformKind].Encode(value) : value;
    }
    catch (error) {
        throw new TransformEncodeError(schema, value, error);
    }
}
function TArray(schema, references, value) {
    const defaulted = Default(schema, value);
    return IsArray(defaulted)
        ? defaulted.map((value) => Visit(schema.items, references, value))
        : defaulted;
}
function TIntersect(schema, references, value) {
    const defaulted = Default(schema, value);
    if (!IsPlainObject(value) || IsValueType(value))
        return defaulted;
    const knownKeys = KeyOfStringResolve(schema);
    const knownProperties = knownKeys.reduce((value, key) => {
        return key in defaulted
            ? { ...value, [key]: Visit(IndexedTypeResolve(schema, [key]), references, value[key]) }
            : value;
    }, defaulted);
    if (!IsTransformType(schema.unevaluatedProperties)) {
        return Default(schema, knownProperties);
    }
    const unknownKeys = Object.getOwnPropertyNames(knownProperties);
    const unevaluatedProperties = schema.unevaluatedProperties;
    return unknownKeys.reduce((value, key) => {
        return !knownKeys.includes(key)
            ? { ...value, [key]: Default(unevaluatedProperties, value[key]) }
            : value;
    }, knownProperties);
}
function TNot(schema, references, value) {
    return Default(schema.not, Default(schema, value));
}
function TObject(schema, references, value) {
    const defaulted = Default(schema, value);
    if (!IsPlainObject(value))
        return defaulted;
    const knownKeys = KeyOfStringResolve(schema);
    const knownProperties = knownKeys.reduce((value, key) => {
        return key in value
            ? { ...value, [key]: Visit(schema.properties[key], references, value[key]) }
            : value;
    }, defaulted);
    if (!IsSchemaType(schema.additionalProperties)) {
        return knownProperties;
    }
    const unknownKeys = Object.getOwnPropertyNames(knownProperties);
    const additionalProperties = schema.additionalProperties;
    return unknownKeys.reduce((value, key) => {
        return !knownKeys.includes(key)
            ? { ...value, [key]: Default(additionalProperties, value[key]) }
            : value;
    }, knownProperties);
}
function TRecord(schema, references, value) {
    const defaulted = Default(schema, value);
    if (!IsPlainObject(value))
        return defaulted;
    const pattern = Object.getOwnPropertyNames(schema.patternProperties)[0];
    const knownKeys = new RegExp(pattern);
    const knownProperties = Object.getOwnPropertyNames(value).reduce((value, key) => {
        return knownKeys.test(key)
            ? { ...value, [key]: Visit(schema.patternProperties[pattern], references, value[key]) }
            : value;
    }, defaulted);
    if (!IsSchemaType(schema.additionalProperties)) {
        return Default(schema, knownProperties);
    }
    const unknownKeys = Object.getOwnPropertyNames(knownProperties);
    const additionalProperties = schema.additionalProperties;
    return unknownKeys.reduce((value, key) => {
        return !knownKeys.test(key)
            ? { ...value, [key]: Default(additionalProperties, value[key]) }
            : value;
    }, knownProperties);
}
function TRef(schema, references, value) {
    const target = Deref(schema, references);
    const resolved = Visit(target, references, value);
    return Default(schema, resolved);
}
function TThis(schema, references, value) {
    const target = Deref(schema, references);
    const resolved = Visit(target, references, value);
    return Default(schema, resolved);
}
function TTuple(schema, references, value) {
    const value1 = Default(schema, value);
    return IsArray(schema.items) ? schema.items.map((schema, index) => Visit(schema, references, value1[index])) : [];
}
function TUnion(schema, references, value) {
    for (const subschema of schema.anyOf) {
        if (!Check(subschema, references, value))
            continue;
        const value1 = Visit(subschema, references, value);
        return Default(schema, value1);
    }
    for (const subschema of schema.anyOf) {
        const value1 = Visit(subschema, references, value);
        if (!Check(schema, references, value1))
            continue;
        return Default(schema, value1);
    }
    return Default(schema, value);
}
function Visit(schema, references, value) {
    const references_ = typeof schema.$id === 'string' ? [...references, schema] : references;
    const schema_ = schema;
    switch (schema[Kind]) {
        case 'Array':
            return TArray(schema_, references_, value);
        case 'Intersect':
            return TIntersect(schema_, references_, value);
        case 'Not':
            return TNot(schema_, references_, value);
        case 'Object':
            return TObject(schema_, references_, value);
        case 'Record':
            return TRecord(schema_, references_, value);
        case 'Ref':
            return TRef(schema_, references_, value);
        case 'This':
            return TThis(schema_, references_, value);
        case 'Tuple':
            return TTuple(schema_, references_, value);
        case 'Union':
            return TUnion(schema_, references_, value);
        default:
            return Default(schema_, value);
    }
}
export function Encode(schema, references, value) {
    return Visit(schema, references, value);
}
