"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Default = void 0;
const index_1 = require("../guard/index");
const type_1 = require("../../type/guard/type");
const index_2 = require("../check/index");
const index_3 = require("../deref/index");
const index_4 = require("../../type/symbols/index");
function ValueOrDefault(schema, value) {
    return !(value === undefined) || !('default' in schema) ? value : schema.default;
}
function IsCheckable(schema) {
    return (0, type_1.TSchema)(schema) && schema[index_4.Kind] !== 'Unsafe';
}
function IsDefaultSchema(value) {
    return (0, type_1.TSchema)(value) && 'default' in value;
}
function TArray(schema, references, value) {
    const defaulted = ValueOrDefault(schema, value);
    if (!(0, index_1.IsArray)(defaulted))
        return defaulted;
    for (let i = 0; i < defaulted.length; i++) {
        defaulted[i] = Visit(schema.items, references, defaulted[i]);
    }
    return defaulted;
}
function TIntersect(schema, references, value) {
    const defaulted = ValueOrDefault(schema, value);
    return schema.allOf.reduce((acc, schema) => {
        const next = Visit(schema, references, defaulted);
        return (0, index_1.IsObject)(next) ? { ...acc, ...next } : next;
    }, {});
}
function TObject(schema, references, value) {
    const defaulted = ValueOrDefault(schema, value);
    if (!(0, index_1.IsObject)(defaulted))
        return defaulted;
    const additionalPropertiesSchema = schema.additionalProperties;
    const knownPropertyKeys = Object.getOwnPropertyNames(schema.properties);
    for (const key of knownPropertyKeys) {
        if (!IsDefaultSchema(schema.properties[key]))
            continue;
        defaulted[key] = Visit(schema.properties[key], references, defaulted[key]);
    }
    if (!IsDefaultSchema(additionalPropertiesSchema))
        return defaulted;
    for (const key of Object.getOwnPropertyNames(defaulted)) {
        if (knownPropertyKeys.includes(key))
            continue;
        defaulted[key] = Visit(additionalPropertiesSchema, references, defaulted[key]);
    }
    return defaulted;
}
function TRecord(schema, references, value) {
    const defaulted = ValueOrDefault(schema, value);
    if (!(0, index_1.IsObject)(defaulted))
        return defaulted;
    const additionalPropertiesSchema = schema.additionalProperties;
    const [propertyKeyPattern, propertySchema] = Object.entries(schema.patternProperties)[0];
    const knownPropertyKey = new RegExp(propertyKeyPattern);
    for (const key of Object.getOwnPropertyNames(defaulted)) {
        if (!(knownPropertyKey.test(key) && IsDefaultSchema(propertySchema)))
            continue;
        defaulted[key] = Visit(propertySchema, references, defaulted[key]);
    }
    if (!IsDefaultSchema(additionalPropertiesSchema))
        return defaulted;
    for (const key of Object.getOwnPropertyNames(defaulted)) {
        if (knownPropertyKey.test(key))
            continue;
        defaulted[key] = Visit(additionalPropertiesSchema, references, defaulted[key]);
    }
    return defaulted;
}
function TRef(schema, references, value) {
    return Visit((0, index_3.Deref)(schema, references), references, ValueOrDefault(schema, value));
}
function TThis(schema, references, value) {
    return Visit((0, index_3.Deref)(schema, references), references, value);
}
function TTuple(schema, references, value) {
    const defaulted = ValueOrDefault(schema, value);
    if (!(0, index_1.IsArray)(defaulted) || (0, index_1.IsUndefined)(schema.items))
        return defaulted;
    const [items, max] = [schema.items, Math.max(schema.items.length, defaulted.length)];
    for (let i = 0; i < max; i++) {
        if (i < items.length)
            defaulted[i] = Visit(items[i], references, defaulted[i]);
    }
    return defaulted;
}
function TUnion(schema, references, value) {
    const defaulted = ValueOrDefault(schema, value);
    for (const inner of schema.anyOf) {
        const result = Visit(inner, references, defaulted);
        if (IsCheckable(inner) && (0, index_2.Check)(inner, result)) {
            return result;
        }
    }
    return defaulted;
}
function Visit(schema, references, value) {
    const references_ = (0, index_1.IsString)(schema.$id) ? [...references, schema] : references;
    const schema_ = schema;
    switch (schema_[index_4.Kind]) {
        case 'Array':
            return TArray(schema_, references_, value);
        case 'Intersect':
            return TIntersect(schema_, references_, value);
        case 'Object':
            return TObject(schema_, references_, value);
        case 'Record':
            return TRecord(schema_, references_, value);
        case 'Ref':
            return TRef(schema_, references_, value);
        case 'This':
            return TThis(schema_, references_, value);
        case 'Tuple':
            return TTuple(schema_, references_, value);
        case 'Union':
            return TUnion(schema_, references_, value);
        default:
            return ValueOrDefault(schema_, value);
    }
}
function Default(...args) {
    return args.length === 3 ? Visit(args[0], args[1], args[2]) : Visit(args[0], [], args[1]);
}
exports.Default = Default;
