"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Encode = exports.TransformEncodeError = exports.TransformEncodeCheckError = void 0;
const type_1 = require("../../type/guard/type");
const index_1 = require("../../type/symbols/index");
const index_2 = require("../guard/index");
const index_3 = require("../../type/keyof/index");
const index_4 = require("../../type/indexed/index");
const index_5 = require("../deref/index");
const index_6 = require("../check/index");
class TransformEncodeCheckError extends Error {
    constructor(schema, value, error) {
        super(`Unable to encode due to invalid value`);
        this.schema = schema;
        this.value = value;
        this.error = error;
    }
}
exports.TransformEncodeCheckError = TransformEncodeCheckError;
class TransformEncodeError extends Error {
    constructor(schema, value, error) {
        super(`${error instanceof Error ? error.message : 'Unknown error'}`);
        this.schema = schema;
        this.value = value;
    }
}
exports.TransformEncodeError = TransformEncodeError;
function Default(schema, value) {
    try {
        return (0, type_1.TTransform)(schema) ? schema[index_1.TransformKind].Encode(value) : value;
    }
    catch (error) {
        throw new TransformEncodeError(schema, value, error);
    }
}
function TArray(schema, references, value) {
    const defaulted = Default(schema, value);
    return (0, index_2.IsArray)(defaulted)
        ? defaulted.map((value) => Visit(schema.items, references, value))
        : defaulted;
}
function TIntersect(schema, references, value) {
    const defaulted = Default(schema, value);
    if (!(0, index_2.IsPlainObject)(value) || (0, index_2.IsValueType)(value))
        return defaulted;
    const knownKeys = (0, index_3.KeyOfStringResolve)(schema);
    const knownProperties = knownKeys.reduce((value, key) => {
        return key in defaulted
            ? { ...value, [key]: Visit((0, index_4.IndexedTypeResolve)(schema, [key]), references, value[key]) }
            : value;
    }, defaulted);
    if (!(0, type_1.TTransform)(schema.unevaluatedProperties)) {
        return Default(schema, knownProperties);
    }
    const unknownKeys = Object.getOwnPropertyNames(knownProperties);
    const unevaluatedProperties = schema.unevaluatedProperties;
    return unknownKeys.reduce((value, key) => {
        return !knownKeys.includes(key)
            ? { ...value, [key]: Default(unevaluatedProperties, value[key]) }
            : value;
    }, knownProperties);
}
function TNot(schema, references, value) {
    return Default(schema.not, Default(schema, value));
}
function TObject(schema, references, value) {
    const defaulted = Default(schema, value);
    if (!(0, index_2.IsPlainObject)(value))
        return defaulted;
    const knownKeys = (0, index_3.KeyOfStringResolve)(schema);
    const knownProperties = knownKeys.reduce((value, key) => {
        return key in value
            ? { ...value, [key]: Visit(schema.properties[key], references, value[key]) }
            : value;
    }, defaulted);
    if (!(0, type_1.TSchema)(schema.additionalProperties)) {
        return knownProperties;
    }
    const unknownKeys = Object.getOwnPropertyNames(knownProperties);
    const additionalProperties = schema.additionalProperties;
    return unknownKeys.reduce((value, key) => {
        return !knownKeys.includes(key)
            ? { ...value, [key]: Default(additionalProperties, value[key]) }
            : value;
    }, knownProperties);
}
function TRecord(schema, references, value) {
    const defaulted = Default(schema, value);
    if (!(0, index_2.IsPlainObject)(value))
        return defaulted;
    const pattern = Object.getOwnPropertyNames(schema.patternProperties)[0];
    const knownKeys = new RegExp(pattern);
    const knownProperties = Object.getOwnPropertyNames(value).reduce((value, key) => {
        return knownKeys.test(key)
            ? { ...value, [key]: Visit(schema.patternProperties[pattern], references, value[key]) }
            : value;
    }, defaulted);
    if (!(0, type_1.TSchema)(schema.additionalProperties)) {
        return Default(schema, knownProperties);
    }
    const unknownKeys = Object.getOwnPropertyNames(knownProperties);
    const additionalProperties = schema.additionalProperties;
    return unknownKeys.reduce((value, key) => {
        return !knownKeys.test(key)
            ? { ...value, [key]: Default(additionalProperties, value[key]) }
            : value;
    }, knownProperties);
}
function TRef(schema, references, value) {
    const target = (0, index_5.Deref)(schema, references);
    const resolved = Visit(target, references, value);
    return Default(schema, resolved);
}
function TThis(schema, references, value) {
    const target = (0, index_5.Deref)(schema, references);
    const resolved = Visit(target, references, value);
    return Default(schema, resolved);
}
function TTuple(schema, references, value) {
    const value1 = Default(schema, value);
    return (0, index_2.IsArray)(schema.items) ? schema.items.map((schema, index) => Visit(schema, references, value1[index])) : [];
}
function TUnion(schema, references, value) {
    for (const subschema of schema.anyOf) {
        if (!(0, index_6.Check)(subschema, references, value))
            continue;
        const value1 = Visit(subschema, references, value);
        return Default(schema, value1);
    }
    for (const subschema of schema.anyOf) {
        const value1 = Visit(subschema, references, value);
        if (!(0, index_6.Check)(schema, references, value1))
            continue;
        return Default(schema, value1);
    }
    return Default(schema, value);
}
function Visit(schema, references, value) {
    const references_ = typeof schema.$id === 'string' ? [...references, schema] : references;
    const schema_ = schema;
    switch (schema[index_1.Kind]) {
        case 'Array':
            return TArray(schema_, references_, value);
        case 'Intersect':
            return TIntersect(schema_, references_, value);
        case 'Not':
            return TNot(schema_, references_, value);
        case 'Object':
            return TObject(schema_, references_, value);
        case 'Record':
            return TRecord(schema_, references_, value);
        case 'Ref':
            return TRef(schema_, references_, value);
        case 'This':
            return TThis(schema_, references_, value);
        case 'Tuple':
            return TTuple(schema_, references_, value);
        case 'Union':
            return TUnion(schema_, references_, value);
        default:
            return Default(schema_, value);
    }
}
function Encode(schema, references, value) {
    return Visit(schema, references, value);
}
exports.Encode = Encode;
